#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <asm/types.h>          /* for videodev2.h */
#include <time.h>
#include "MyCapture.h"


#define CLEAR(x) memset (&(x), 0, sizeof (x))
static FILE* before = NULL, *after = NULL;
int MyVideoCapture::m_fd = -1;
void** MyVideoCapture::m_startArray = NULL;
size_t MyVideoCapture::m_length = 0;
size_t MyVideoCapture::m_bufferNumber=0;

MyVideoCapture::MyVideoCapture()
{
    m_pixelFormat = V4L2_PIX_FMT_YUYV;
    m_width = 0;
    m_height = 0;
    m_startArray = NULL;
    m_length = 0;
    m_bufferNumber = 1;
}

MyVideoCapture::~MyVideoCapture()
{
    close_device();
}

bool MyVideoCapture::init(unsigned long width, unsigned long height)
{
    m_width = width;
    m_height = height;
    return open_device();
}


bool MyVideoCapture::open_device()
{
    for (int i = 0; i < 64; i ++)
    {
        struct stat st;
        sprintf(m_dev_name, "/dev/video%d", i);
        if (0 == stat (m_dev_name, &st))
        {
            if (S_ISCHR (st.st_mode))
            {
                if ((m_fd = open(m_dev_name, O_RDWR | O_NONBLOCK, 0)) != -1)
                {
                    if (init_device())
                    {
                        return true;
                    }
                    close(m_fd);
                }
            }
        }
    }
    return false;
}

int MyVideoCapture::xioctl(int fd, int request, void* arg)
{
    int r;
    do
    {
        r = ioctl (fd, request, arg);
    }
    while ((-1 == r && EINTR == errno) || (-1 == r && EAGAIN == errno));
    if (r == -1)
    {
        perror("ioctl returns -1");
    }

    return r;
}

bool MyVideoCapture::listAllFormats()
{
    v4l2_fmtdesc desc = {0};
    desc.index = 0;
    desc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    while (xioctl(m_fd, VIDIOC_ENUM_FMT, &desc) != -1)
    {
        printf("%d: %s\n", desc.index, desc.description);
        desc.index ++;
    }

    // enum video input
    struct v4l2_input input = {0};
    input.index = 0;
    while (xioctl(m_fd, VIDIOC_ENUMINPUT, &input) != -1)
    {
        printf("name: %s\n type: %u\n std: %u\n status: %u\n capabilities: %u\n",
               input.name, input.type, input.std, input.status, input.capabilities);
        input.index ++;
    }
    return true;
}


bool MyVideoCapture::init_device()
{
    struct v4l2_capability cap;

    if (-1 == xioctl (m_fd, VIDIOC_QUERYCAP, &cap))
    {
        return false;
    }
    printf("######################capabilities : %u##############################\n", cap.capabilities);
    if (cap.capabilities & V4L2_CAP_DEVICE_CAPS)
    {
        printf("######################device_caps : %u##############################\n", cap.device_caps);
        cap.capabilities = cap.device_caps;
    }
    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) ||
            !(cap.capabilities & (V4L2_CAP_STREAMING | V4L2_CAP_READWRITE)))
    {
        return false;
    }
    struct v4l2_format fmt = { 0, };
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (xioctl(m_fd, VIDIOC_G_FMT, &fmt) == -1)
    {
        return false;
    }
    struct v4l2_streamparm parm = { 0, };
    parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (xioctl(m_fd, VIDIOC_G_PARM, &parm) == -1)
    {
        return false;
    }
    else
    {
        struct v4l2_captureparm& captureParm = parm.parm.capture;
        printf("__u32 capability; Supported modes: %d\n", captureParm.capability);
        printf("__u32 capturemode; Current mode: %d\n", captureParm.capturemode);
        printf("struct v4l2_fract timeperframe; Time per frame in seconds:\n");
        printf("__u32 numerator; %d\n", captureParm.timeperframe.numerator);
        printf("__u32 denominator; %d\n", captureParm.timeperframe.denominator);
        printf("__u32 extendedmode; Driver-specific extensions: %d\n", captureParm.extendedmode);
        printf("__u32 readbuffers; # of buffers for read: %d\n", captureParm.readbuffers);
    }

    v4l2_fmtdesc desc;
    desc.index = 0;
    desc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    while (xioctl(m_fd, VIDIOC_ENUM_FMT, &desc) != -1)
    {
        if (desc.pixelformat == m_pixelFormat)
        {
            CLEAR (m_fmt);
            m_fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            m_fmt.fmt.pix.pixelformat = m_pixelFormat;

            m_fmt.fmt.pix.width = m_width;
            m_fmt.fmt.pix.height = m_height;
            if (-1 == xioctl (m_fd, VIDIOC_S_FMT, &m_fmt))
            {
                return false;
            }
            else
            {
                return true; // init_buffer();
            }

        }
        desc.index ++;
    }
    return false;
}

static void printBufStatus(struct v4l2_buffer& buf)
{
    printf("****************printBufStatus***************\n\n");
    printf("__u32 index; %d\n", buf.index);
    printf("__u32 type; %d\n", buf.type);
    printf("__u32 bytesused; %d\n", buf.bytesused);

    printf("__u32 field; %d\n", buf.field);
    //V4L2_BUF_FLAG_MAPPED, V4L2_BUF_FLAG_PREPARED, V4L2_BUF_FLAG_QUEUED and V4L2_BUF_FLAG_DONE
    switch (buf.flags)
    {
    case V4L2_BUF_FLAG_MAPPED:
        printf("__u32 flags; V4L2_BUF_FLAG_MAPPED\n");
        break;
    case V4L2_BUF_FLAG_PREPARED:
        printf("__u32 flags; V4L2_BUF_FLAG_PREPARED\n");
        break;
    case V4L2_BUF_FLAG_QUEUED:
        printf("__u32 flags; V4L2_BUF_FLAG_QUEUED\n");
        break;
    case V4L2_BUF_FLAG_DONE:
        printf("__u32 flags; V4L2_BUF_FLAG_DONE\n");
        break;
    case V4L2_BUF_FLAG_KEYFRAME:
        printf("__u32 flags; V4L2_BUF_FLAG_KEYFRAME\n");
        break;
    case V4L2_BUF_FLAG_PFRAME:
        printf("__u32 flags; V4L2_BUF_FLAG_PFRAME\n");
        break;
    case V4L2_BUF_FLAG_BFRAME:
        printf("__u32 flags; V4L2_BUF_FLAG_BFRAME\n");
        break;

    case V4L2_BUF_FLAG_ERROR:
        printf("__u32 flags; V4L2_BUF_FLAG_ERROR\n");
        break;
    case V4L2_BUF_FLAG_TIMECODE:
        printf("__u32 flags; V4L2_BUF_FLAG_TIMECODE\n");
        break;
    case V4L2_BUF_FLAG_TIMESTAMP_MASK:
        printf("__u32 flags; V4L2_BUF_FLAG_TIMESTAMP_MASK\n");
        break;
    case V4L2_BUF_FLAG_TIMESTAMP_UNKNOWN:
        printf("__u32 flags; V4L2_BUF_FLAG_TIMESTAMP_UNKNOWN\n");
        break;
    case V4L2_BUF_FLAG_NO_CACHE_INVALIDATE:
         printf("__u32 flags; V4L2_BUF_FLAG_NO_CACHE_INVALIDATE\n");
         break;
    case V4L2_BUF_FLAG_NO_CACHE_CLEAN:
         printf("__u32 flags; V4L2_BUF_FLAG_NO_CACHE_CLEAN\n");
         break;

    case V4L2_BUF_FLAG_TIMESTAMP_MONOTONIC:
        printf("__u32 flags; V4L2_BUF_FLAG_TIMESTAMP_MONOTONIC\n");
        break;
    case V4L2_BUF_FLAG_TIMESTAMP_COPY:
        printf("__u32 flags; V4L2_BUF_FLAG_TIMESTAMP_COPY\n");
        break;
    default:
        printf("__u32 flags; *****unknown flags: %d!!!!*****\n", buf.flags);
        break;

    }


    printf("struct timeval timestamp; %u, %u\n", buf.timestamp.tv_sec, buf.timestamp.tv_usec);
    printf("struct v4l2_timecode timecode; \n");
    printf("__u32 type; %d\n", buf.timecode.type);
    printf("__u32 flags; %d\n", buf.timecode.flags);
    printf("__u8 frames; %u\n", buf.timecode.frames);
    printf("__u8 seconds; %u\n", buf.timecode.seconds);
    printf("__u8 minutes; %u\n", buf.timecode.minutes);
    printf("__u8 hours; %u\n", buf.timecode.hours);
    printf("__u8  userbits[4]; %u%u%u%u\n", buf.timecode.userbits[0],
           buf.timecode.userbits[1], buf.timecode.userbits[2], buf.timecode.userbits[3]);

    printf("__u32 sequence; %u\n", buf.sequence);

    /* memory location */
    printf("__u32 memory; %u\n", buf.memory);
    printf("__u32 offset: %u\n", buf.m.offset);
//    union {
//        __u32           offset;
//        unsigned long   userptr;
//        struct v4l2_plane *planes;
//        __s32       fd;
//    } m;
    printf("__u32 length; %d\n", buf.length);
//    __u32           reserved2;
//    __u32           reserved;
    printf("****************printBufStatus***************\n\n");
}
bool MyVideoCapture::init_buffer()
{
    struct v4l2_requestbuffers req;
    m_startArray = new void*[m_bufferNumber];
    if (m_startArray == NULL)
    {
        return false;
    }
    memset(m_startArray, 0, sizeof(void*)*m_bufferNumber);

    CLEAR(req);
    req.count = m_bufferNumber;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_USERPTR;
    if (xioctl(m_fd, VIDIOC_REQBUFS, &req) == -1)
    {
        return false;
    }
    printf("VIDIOC_REQBUFS returns number of buffer %d\n", req.count);

    m_bufferNumber = req.count;
    struct v4l2_buffer buf;

    bool bResult = true;
    for (size_t i = 0; i < req.count; i ++)
    {
        CLEAR(buf);
        buf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index  = i;
        buf.field = V4L2_FIELD_NONE;

        if (xioctl(m_fd, VIDIOC_QUERYBUF, &buf) == -1)
        {
            bResult = false;
            break;
        }
        printf("************************before qbuf************************\n");
        printBufStatus(buf);
        m_length = buf.length;
        //m_startArray[i] = mmap(NULL, buf.length, PROT_READ|PROT_WRITE, MAP_SHARED, m_fd, buf.m.offset);
        m_startArray[i] = malloc(m_length);

        if (NULL == m_startArray[i])
        {
            bResult = false;
            break;
        }
        buf.flags = 0;
        buf.reserved = buf.reserved2 = 0;
        printf("************************before qbuf************************\n");
        buf.m.userptr = (unsigned long)m_startArray[i];
        if (xioctl(m_fd, VIDIOC_QBUF, &buf) == -1)
        {
            bResult = false;
            break;
        }
        else
        {
            if (xioctl(m_fd, VIDIOC_QUERYBUF, &buf) == -1)
            {
                bResult = false;
                break;
            }
            printf("************************after qbuf************************\n");
            printBufStatus(buf);
            printf("************************after qbuf************************\n");
        }
    }
    if (bResult)
    {
        enum v4l2_buf_type type;
        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if (xioctl(m_fd, VIDIOC_STREAMON, &type) == -1)
        {
            return false;
        }
    }
    return bResult;
}

void MyVideoCapture::close_device()
{
    enum v4l2_buf_type type;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    xioctl(m_fd, VIDIOC_STREAMOFF, &type);
    if (m_startArray != NULL)
    {
        for (size_t i = 0; i < m_bufferNumber; i ++)
        {
            //munmap(m_startArray[i], m_length);
            free(m_startArray[i]);
        }
        delete []m_startArray;
        m_startArray = NULL;
        m_length = 0;
    }
    if (m_fd != -1)
    {
        close(m_fd);
        m_fd = -1;
        if (before)
        {
            fclose(before);
        }
        if (after)
        {
            fclose(after);
        }
    }
}

//void MyVideoCapture::startCaptureAndDisplay(unsigned long pixelFormat,
//               unsigned long width, unsigned long height, size_t bufferNumber)
//{
////    if (m_fd == -1)
////    {
////        m_pixelFormat = pixelFormat;
////        m_width = width;
////        m_height = height;
////        m_bufferNumber = bufferNumber;
////        if (open_device())
////        {
////            if (pthread_create(&m_pthread, NULL, &MyVideoCapture::mainloop, NULL) == 0)
////            {
////                pthread_join(m_pthread, NULL);
////                close_device();
////            }
////        }
////        else
////        {
////            close_device();
////        }
////    }
//}
//
//size_t MyVideoCapture::startCapture(unsigned long pixelFormat,
//               unsigned long width, unsigned long height, size_t bufferNumber)
//{
//    size_t size = 0;
//    if (m_fd == -1)
//    {
//        m_pixelFormat = pixelFormat;
//        m_width = width;
//        m_height = height;
//        m_bufferNumber = bufferNumber;
//        if (open_device())
//        {
//            size = m_length;
//            before = fopen("before", "w");
//            after = fopen("after", "w");
//        }
//        else
//        {
//            close_device();
//
//        }
//    }
//    return size;
//}
//
//bool MyVideoCapture::captureFrame(unsigned char*dstPtr, bool bSkip)
//{
////    int r;
////    do
////    {
////        struct timeval tv;
////        fd_set fds;
////        FD_ZERO(&fds);
////        FD_SET(m_fd, &fds);
////        /* Timeout. */
////        tv.tv_sec = 1;
////        tv.tv_usec = 0;
////
////        r = select(m_fd, &fds, NULL, NULL, &tv);
////    }
////    while ((r == -1 && (errno = EINTR)));
////    if (r == -1)
////    {
////        perror("select returns -1");
////        return false;
////    }
//
//    struct v4l2_buffer buf;
//    CLEAR(buf);
//    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//    buf.memory = V4L2_MEMORY_USERPTR;
//    buf.index = 0;
//    if (xioctl(m_fd, VIDIOC_QUERYBUF, &buf) == -1)
//    {
//        return false;
//    }
//    printBufStatus(buf);
//
//    printf("************************before dqbuf************************\n");
//    if (xioctl(m_fd, VIDIOC_DQBUF, &buf) == -1)
//    {
//        return false;
//    }
//    unsigned char* srcPtr = (unsigned char*)m_startArray[buf.index];
//    if (!bSkip)
//    {
//        memcpy(dstPtr, srcPtr, m_length);
//    }
//    printf("************************before qbuf************************\n");
//    if (xioctl(m_fd, VIDIOC_QBUF, &buf) == -1)
//    {
//        return false;
//    }
//    return true;
//}


//bool MyVideoCapture::captureFrame(unsigned char*dstPtr, bool bSkip)
//{
//    int size = read(m_fd, dstPtr, m_length);
//    if (size != m_length)
//    {
//        printf("######################read return %d########################\n", size);
//        return false;
//    }
//    return true;
//}

bool MyVideoCapture::captureFrame(unsigned char*dstPtr[4], bool bSkip)
{
    int r;
    do
    {
        struct timeval tv;
        fd_set fds;
        FD_ZERO(&fds);
        FD_SET(m_fd, &fds);
        /* Timeout. */
        tv.tv_sec = 1;
        tv.tv_usec = 0;

        r = select(m_fd + 1, &fds, NULL, NULL, &tv);
    }
    while ((r == -1 && (errno = EINTR)));
    if (r == -1)
    {
        return false;
    }
    struct v4l2_buffer buf;
    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_USERPTR;
    if (xioctl(m_fd, VIDIOC_DQBUF, &buf) == -1)
    {
        return false;
    }
    if (!bSkip)
    {
        unsigned char* srcPtr = (unsigned char*)m_startArray[buf.index];

        unsigned char* yPtr = dstPtr[0], *uPtr=dstPtr[1], *vPtr= dstPtr[2];

        //fwrite(srcPtr, m_length, 1, before);

        //fwrite(dstPtr[0], m_width*m_height, 1, after);
        //fwrite(dstPtr[1], m_width*m_height/4, 1, after);
        //fwrite(dstPtr[2], m_width*m_height/4, 1, after);
    }

    if (xioctl(m_fd, VIDIOC_QBUF, &buf) == -1)
    {
        return false;
    }
    return true;
}


//void* MyVideoCapture::mainloop(void* arg)
//{
//    MySDLDisplay display;
//    SDL_Event event;
//    struct timeval last, now;
//    int counter = 0;
//    if (!display.init())
//    {
//        return NULL;
//    }
//    while (true)
//    {
//        if (counter == 0 )
//        {
//            gettimeofday(&last, NULL);
//        }
//
//        SDL_PollEvent(&event);
//        if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE)
//        {
//            break;
//        }
//        int r;
//        do
//        {
//            struct timeval tv;
//            fd_set fds;
//            FD_ZERO(&fds);
//            FD_SET(m_fd, &fds);
//            /* Timeout. */
//            tv.tv_sec = 1;
//            tv.tv_usec = 0;
//
//            r = select(m_fd + 1, &fds, NULL, NULL, &tv);
//        }
//        while ((r == -1 && (errno = EINTR)));
//        if (r == -1)
//        {
//            perror("select");
//            break;
//        }
//        struct v4l2_buffer buf;
//        CLEAR(buf);
//        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//        buf.memory = V4L2_MEMORY_MMAP;
//        if (xioctl(m_fd, VIDIOC_DQBUF, &buf) == -1)
//        {
//            break;
//        }
//
//        if (!display.displayImage((unsigned char*)m_startArray[buf.index]))
//        {
//            break;
//        }
//
//        if (xioctl(m_fd, VIDIOC_QBUF, &buf) == -1)
//        {
//            break;
//        }
//        counter ++;
//        if (counter % FPS_COUNT_NUMBER == 0)
//        {
//            gettimeofday(&now, NULL);
//            double diff = (now.tv_sec - last.tv_sec) + (double)(now.tv_usec - last.tv_usec)/1000000.0;
//            counter = 0;
//            printf("fps:%f\n", (double)FPS_COUNT_NUMBER/diff);
//
//        }
//    }
//
//    return NULL;
//}

/*
void* MyVideoCapture::mainloop2(void* arg)
{
    MySDLDisplay display;
    SDL_Event event;
    struct timeval last, now;
    int counter = 0;
    if (!display.init())
    {
        return NULL;
    }
    while (true)
    {
        if (counter == 0 )
        {
            gettimeofday(&last, NULL);
        }

        SDL_PollEvent(&event);
        if (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE)
        {
            break;
        }
        int r;
        do
        {
            struct timeval tv;
            fd_set fds;
            FD_ZERO(&fds);
            FD_SET(m_fd, &fds);
            // Timeout.
            tv.tv_sec = 1;
            tv.tv_usec = 0;

            r = select(m_fd + 1, &fds, NULL, NULL, &tv);
        }
        while ((r == -1 && (errno = EINTR)));
        if (r == -1)
        {
            perror("select");
            break;
        }
        struct v4l2_buffer buf;
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        if (xioctl(m_fd, VIDIOC_DQBUF, &buf) == -1)
        {
            break;
        }

        if (!display.displayImage((unsigned char*)m_startArray[buf.index]))
        {
            break;
        }

        if (xioctl(m_fd, VIDIOC_QBUF, &buf) == -1)
        {
            break;
        }
        counter ++;
        if (counter % FPS_COUNT_NUMBER == 0)
        {
            gettimeofday(&now, NULL);
            double diff = (now.tv_sec - last.tv_sec) + (double)(now.tv_usec - last.tv_usec)/1000000.0;
            counter = 0;
            printf("fps:%f\n", (double)FPS_COUNT_NUMBER/diff);

        }
    }

    return NULL;
}
*/

/*
int main()
{
    MyVideoCapture* ptr = MyVideoCapture::createInstance();
    void* pStart = NULL;
    size_t length = 0;
    if (ptr->start(&pStart, &length))
    {
        ptr->end();
    }
    return 0;

}


bool MyVideoCapture::captureFrame(unsigned char*dstPtr[4], bool bSkip)
{
    int r;
    do
    {
        struct timeval tv;
        fd_set fds;
        FD_ZERO(&fds);
        FD_SET(m_fd, &fds);

        tv.tv_sec = 1;
        tv.tv_usec = 0;

        r = select(m_fd + 1, &fds, NULL, NULL, &tv);
    }
    while ((r == -1 && (errno = EINTR)));
    if (r == -1)
    {
        return false;
    }
    struct v4l2_buffer buf;
    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if (xioctl(m_fd, VIDIOC_DQBUF, &buf) == -1)
    {
        return false;
    }
    unsigned char* srcPtr = (unsigned char*)m_startArray[buf.index];
    if (!bSkip)
    {
        size_t planeSize = m_width*m_height;
        unsigned char* yPtr = dstPtr[0], *uPtr=dstPtr[1], *vPtr= dstPtr[2];
        bool bEvenRow = true;
        for (size_t r = 0; r < m_height; r ++)
        {
            bool bEvenCol = true;
            for (size_t c = 0; c < m_width; c ++)
            {
                *yPtr = *srcPtr++;
                if (bEvenRow)
                {
                    if (bEvenCol)
                    {
                        bEvenCol = false;
                        *uPtr++ = *srcPtr++;
                        *vPtr++ = *srcPtr++;
                    }
                    else
                    {
                        bEvenCol = true;
                    }
                }
                else
                {
                    if (bEvenCol)
                    {
                        bEvenCol = false;
                        srcPtr++;
                        srcPtr++;
                    }
                    else
                    {
                        bEvenCol = true;
                    }
                }
            }
            if (bEvenRow)
            {
                bEvenRow = false;
            }
            else
            {
                bEvenRow = true;
            }
        }
    }

    if (xioctl(m_fd, VIDIOC_QBUF, &buf) == -1)
    {
        return false;
    }
    return true;
}
*/
